from mautil.tf.gan import GAN
from mautil.tf.nn import NN


class BigGANBlock(NN):
    def __init__(self, name='BigGANBlock' ):
        pass

    def call(self, inputs, training=None, mask=None):
        pass


class BigGAN(GAN):
    cfg = GAN.cfg.copy()



__version__ = "unknown"
try:
    from ._version import __version__
except ImportError:
    raise Exception("version not found")

from .util import *
from .batch_reader import BatchReader
from .basic_models import BasicCFG, CFG, Model, InputFeature
from .text_process import Vocabulary



import logging, os, json
import random
from glob import glob
from collections import OrderedDict, defaultdict
import numpy as np
import inspect
from functools import partial
from tqdm import tqdm
from copy import deepcopy
import torch
import torch.nn.functional as F

try:
    import torch_xla.core.xla_model as xm, torch_xla.distributed.parallel_loader as pl, torch_xla.distributed.xla_multiprocessing as xmp
except Exception as e:
    pass

try:
    import torchcontrib
except:
    pass

from mautil.basic_models import CFG as BasicCFG, Model as BasicModel, LossHist
from mautil.pt import util as pt_util
from mautil import util

WEIGHT_SUFFIX = '.bin'


logger = logging.getLogger(__name__)


class CFG(BasicCFG):
    def __init__(self):
        super(CFG, self).__init__()
        self.loss = 'mse'
        self.global_step = None
        self.gradient_clip = None
        self.keep_checkpoint_every_n_hours = 1000000000
        self.save_keep = 1000000000
        self.save_summary_step = 1000000000
        self.use_tpu = False
        self.use_tffit = True
        self.opt = 'torch.optim.AdamW'
        self.tpu_name = None
        self.warm_start_path = None  # used by tpu
        self.tpu_zone = None
        self.gcp_project = None
        self.num_core_per_host = 1
        self.num_hosts = 1
        self.tpu_loop_iterations = 2
        self.run_eagerly = False
        self.accumulated_batch_size = 1
        self.compile_model = False
        # xla
        self.xla_procs = 8
        self.xla_spawn_method = 'fork'

        # dataset
        self.n_dl_worker = 0

        # swa
        self.swa_batch_size = None


#    def dump(self, fpath):
#        os.makedirs(os.path.dirname(fpath), exist_ok=True)
#        with open(fpath, 'w') as f:
#            json.dump(self.__dict__, f)


class XLA(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, model, owner):
        def wrapper(*args, **kwargs):
            kwargs = util.fill_kwargs(self.f, args, kwargs)
            kwargs['self'] = model
            if model.cfg.use_tpu:
                xla_procs = kwargs.get('xla_procs', model.cfg.xla_procs)
                if xla_procs>1:
                    if model.cfg.xla_share_model:
                        kwargs['pt_model'] = xmp.MpModelWrapper(model.create_core_model())
                    return xmp.spawn(self.run, args=(model, kwargs), nprocs=xla_procs, start_method=model.cfg.xla_spawn_method)
                else:
                    kwargs = self.process_args(model, kwargs)
                    return self.f(**kwargs)
            else:
                return self.f(**kwargs)
        return wrapper

    def process_args(self, model, kwargs):
        if self.f.__name__ == 'predict':
            ds, data_type = kwargs.get('ds'), kwargs.get('data_type')
            device = xm.xla_device()
            ds = model._gen_distribute_dl(ds, data_type)
            #ds = pl.ParallelLoader(ds, [device]).per_device_loader(device)
            kwargs['ds'] = ds
        else:
            # fit func
            train_ds, val_ds = kwargs.get('train_ds'), kwargs.get('val_ds')
            if train_ds is not None:
                train_ds = model._gen_distribute_dl(train_ds, 'train')
                kwargs['train_ds'] = train_ds
            if val_ds is not None:
                val_ds = model._gen_distribute_dl(val_ds, 'val')
                kwargs['val_ds'] = val_ds
        return kwargs

    def run(self, index, model, kwargs):
        kwargs = self.process_args(model, kwargs)
        return self.f(**kwargs, index=index)


def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True

"""
swa related. should remove in future
"""
# BatchNorm utils
def _check_bn_apply(module, flag):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        flag[0] = True


def _check_bn(model):
    flag = [False]
    model.apply(lambda module: _check_bn_apply(module, flag))
    return flag[0]


def _reset_bn(module):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        module.running_mean = torch.zeros_like(module.running_mean)
        module.running_var = torch.ones_like(module.running_var)


def _get_momenta(module, momenta):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        momenta[module] = module.momentum


def _set_momenta(module, momenta):
    if issubclass(module.__class__, torch.nn.modules.batchnorm._BatchNorm):
        module.momentum = momenta[module]

class PTModel(BasicModel):
    cfg = CFG()

    def __init__(self, name='PTModel', cfg={}):
        super(PTModel, self).__init__(name, cfg)
        self._restored = False
        self._restored_epoch = None
        self._accumulated_grads = None
        self._batch_id = None
        self._opt = None
        self._device = None
        self._is_cuda_available = None
        self._lr_scheduler = None
        self._grad_scaler = None
        self._ema_model = None
        set_seed(self.cfg.seed)

        if self.cfg.restore:
            # restore cfg first
            super(PTModel, self).restore(model_dir=self.cfg.model_dir)

    def create_core_model(self, **kwargs):
        pass

    def get_device(self):
        if self.cfg.use_tpu:
            device = xm.xla_device()
        elif self._is_cuda_available:
            if self.cfg.device is None:
                device = torch.device('cuda')
            else:
                device = torch.device('cuda:'+','.join(self.cfg.device))
        else:
            device = torch.device('cpu')
        return device

    def _create_grad_scaler(self):
        return torch.cuda.amp.GradScaler()

    def to_device(self):
        device = self._device
        if torch.cuda.device_count() > 1 and self.cfg.use_multiple_gpu:
            if self.cfg.device is not None and len(self.cfg.device)>1:
                self._model = nn.DataParallel(self._model, device_ids=self.cfg.device)
                logger.info("use GPUS:", self.cfg.device)
            else:
                self._model = torch.nn.DataParallel(self._model)
                logger.info("use %s GPUS!", torch.cuda.device_count())
        self._model = self._model.to(device)
        if self._ema_model is not None:
            self._ema_model.ema = self._ema_model.ema.to(device)
        logger.info('created model on device:%s', device)

    def create_model(self, opt=None, pt_model=None, **kwargs):
        self._is_cuda_available = torch.cuda.is_available()
        self._device = self.get_device()
        if self.cfg.use_fp16:
            self._grad_scaler = self._create_grad_scaler()
        if pt_model is None:
            self._model = self.create_core_model(**kwargs)
            if self.cfg.ema is not None and self.cfg.ema > 0:
                self._ema_model = ModelEma(self._model, decay=self.cfg.ema)
                logger.info('ema model created')
        else:
            self._model = pt_model

        self.to_device()
        # it is said for some optimizer like Adagrad need create opt after model.cuda()
        if opt is not None:
            self._opt = opt
        else:
            if self.cfg.use_deepspeed:
                self._ds_model, self._ds_opt, self._lr_scheduler = util.init_deepspeed(self.cfg.deepspeed, self._model)
                self._opt = self._ds_opt.optimizer
                self._model = self._ds_model.module
            else:
                self._opt, self._lr_scheduler = self._create_optimizer()

        if self.cfg.restore:
            self.restore(restore_epoch=self.cfg.restore_epoch, model_dir=self.cfg.model_dir)

        if self.cfg.use_tpu and xm.xrt_world_size()>1:
            xm.rendezvous('create model done')

    def _create_lr_scheduler(self, opt):
        if self.cfg.lr_scheduler is None:
            lr_scheduler = None
        elif self.cfg.lr_scheduler =='ld':
            lr_scheduler = pt_util.get_linear_decay_schedule_with_warmup(opt, self.cfg.n_lr_warmup_step, self.cfg.n_lr_decay_step, self.cfg.lr_decay_rate)
        else:
            lr_scheduler = util.dynamic_import(self.cfg.lr_scheduler)(opt, **self.cfg.lr_scheduler_ca_paras)
        return lr_scheduler

    def _create_optimizer(self, parameters=None, lr=None):
        if parameters is None:
            parameters = self._model.parameters()
        if lr is None:
            lr = self.cfg.lr
        # create after model.cuda()
#        if self.cfg.opt == 'adamW':
#            opt = torch.optim.AdamW(parameters, lr=self.cfg.lr, weight_decay=self.cfg.weight_decay)
#        elif self.cfg.opt == 'adam':
#            opt = torch.optim.Adam(parameters, lr=self.cfg.lr)
#        else:
#            opt = util.dynamic_import(self.cfg.opt)(parameters, lr=self.cfg.lr, **self.cfg.opt_paras)
        opt = util.dynamic_import(self.cfg.opt)(filter(lambda p:p.requires_grad, parameters), lr=lr, **self.cfg.opt_paras)
        if self.cfg.swa_start_epoch >0:
            opt = torchcontrib.optim.SWA(opt)

        lr_scheduler = self._create_lr_scheduler(opt)
        return opt, lr_scheduler

    def save(self, global_step=None, save_path=None, epoch=None, save_opt=False, **kwargs):
        if self.cfg.n_keep_ckpt>1:
            self.cfg.saved_epoch = epoch
        else:
            self.cfg.saved_epoch = None
        if save_path is None:
            save_name = 'model-{}{}'.format(epoch, WEIGHT_SUFFIX) if epoch and self.cfg.n_keep_ckpt>1 else 'model'+ WEIGHT_SUFFIX
            if self.cfg.model_dir is None:
                save_path = self.gen_fname(save_name)
            else:
                save_path = os.path.join(self.cfg.model_dir, save_name)
        to_save_model = self._model
        if not self.cfg.save_ema and self._ema_model is not None:
            to_save_model = self._ema_model.ema # switch ema
        if self.cfg.save_half:
            to_save_model = deepcopy(to_save_model).half()
        state_dict = {'epoch': epoch,
                      'global_step': global_step,
                      'model_state_dict': to_save_model.state_dict(),
                      'rng_state':torch.get_rng_state(),
                      }
        if save_opt and not self.cfg.use_deepspeed:
            assert self.cfg.save_ema or self._ema_model is None, "should save ema for resume training"
            state_dict['optimizer_state_dict'] = self._opt.state_dict()
            if self._lr_scheduler is not None:
                state_dict['lr_scheduler_state_dict'] = self._lr_scheduler.state_dict()
        state_dict.update(kwargs)
        if self.cfg.save_ema and self._ema_model is not None:
            state_dict['ema_state_dict'] = self._ema_model.ema.state_dict()

        self.update_saved_ckpt([save_path], epoch)
        if self.cfg.use_tpu:
            if xm.is_master_ordinal():
                super(PTModel, self).save()
        else:
            super(PTModel, self).save()

        if self.cfg.use_tpu:
            xm.save(state_dict, save_path)
        else:
            torch.save(state_dict, save_path)
        if self.cfg.use_deepspeed:
            self._ds_model.save_checkpoint(self.gen_fname())
            logger.info('deepspeed model saved to %s', self.gen_fname())
        self.info("Model saved to file:{}".format(save_path))

    def update_saved_ckpt(self, save_paths, epoch):
        save_paths = [os.path.basename(save_path) for save_path in save_paths] # use rel path, not exposure private info
        self.cfg.saved_ckpts.append([epoch, save_paths])
        if len(self.cfg.saved_ckpts)>self.cfg.n_keep_ckpt:
            for save_path in self.cfg.saved_ckpts[0][1]:
                if os.path.exists(save_path):
                    os.remove(save_path)
            self.cfg.saved_ckpts = self.cfg.saved_ckpts[1:]
            #self.cfg.saved_ckpts = self.cfg.saved_ckpts[-self.cfg.n_keep_ckpt:]

    def get_checkpoint_path_bak(self, model_dir=None, restore_epoch=None):
        if model_dir is None:
            model_dir = self.gen_fname()
        if restore_epoch is None and self.cfg.saved_epoch is not None:
            restore_epoch = self.cfg.saved_epoch
        save_name = 'model-{}{}'.format(restore_epoch, WEIGHT_SUFFIX) if restore_epoch else 'model' + WEIGHT_SUFFIX
        model_path = os.path.join(model_dir, save_name)
        return model_path

    def get_checkpoint_path(self, model_dir=None, restore_epoch=None):
        if restore_epoch is None:
            restore_epoch, save_paths = self.cfg.saved_ckpts[-1]
        else:
            save_paths = None
            for i, (save_epoch, save_paths) in enumerate(self.cfg.saved_ckpts[::-1]):
                if save_epoch == restore_epoch:
                    break
                else:
                    save_paths = None

        if model_dir is None:
            model_dir = self.gen_fname()
        save_paths = [os.path.join(model_dir, save_path) for save_path in save_paths]
        return save_paths, restore_epoch

    def load_state_dict(self, model, state_dict):
        missing_keys, unexpected_keys = model.load_state_dict(state_dict, strict=self.cfg.restore_strict)
        if len(missing_keys) > 0:
            self.info('missing_keys %s', missing_keys)
        if len(unexpected_keys) > 0:
            self.info('unexpected_keys %s', unexpected_keys)


    def restore(self, restore_epoch=None, epoch=None, model_dir=None):
        if self._model is None:
            self.create_model()

        if not self._restored:
            model_paths, restore_epoch = self.get_checkpoint_path(model_dir, restore_epoch)
            model_path = model_paths[0]
            ckpt = torch.load(model_path, map_location=torch.device('cpu'))
            self.load_state_dict(self._model, ckpt['model_state_dict'])
            if self._ema_model is not None and 'ema_state_dict' not in ckpt:
                self.load_state_dict(self._ema_model.ema, ckpt['model_state_dict'])
                self.info("copied ema model weights from model")
            if 'optimizer_state_dict' in ckpt:
                self._opt.load_state_dict(ckpt['optimizer_state_dict'])
                self.info("optimizer state restored")
            if 'lr_scheduler_state_dict' in ckpt:
                self._lr_scheduler.load_state_dict(ckpt['lr_scheduler_state_dict'])
                self.info("lr scheduler state restored")
            if 'ema_state_dict' in ckpt:
                self.info('restor ema model:%s', model_path)
                self.load_state_dict(self._ema_model.ema, ckpt['ema_state_dict'])
                self.info('ema model restored from %s', model_path)
            if 'rng_state' in ckpt:
                torch.set_rng_state(ckpt['rng_state'])
                self.info("torch random state restored")

            if 'epoch' in ckpt:
                self._restored_epoch = ckpt['epoch']
            else:
                self._restored_epoch = restore_epoch

            if self.cfg.use_deepspeed:
                self._ds_model.load_checkpoint(self.gen_fname(), load_optimizer_states=True, load_lr_scheduler_states=True)

            self._restored = True
            self.info('model restored from %s, epoch:%s', model_path, self._restored_epoch)

    def calc_loss(self, inputs, outputs, model=None):
        """

        :param inputs:
        :param outputs:
        :return:  dictionary of losses
        """
        if model is None:
            model = self._model
        if isinstance(model, torch.nn.DataParallel):
            model = model.module

        losses = model.calc_loss(inputs, outputs)
        return losses

    def _step_opt(self, opt, lr_scheduler=None):
        if self.cfg.use_deepspeed:
            self.deepspeed.step()
        else:
            if self.cfg.gradient_clip is not None:
                for gp in opt.param_groups:
                    torch.nn.utils.clip_grad_norm_(gp['params'], self.cfg.gradient_clip)
            if self.cfg.use_tpu:
                xm.optimizer_step(opt)
            elif self.cfg.use_fp16:
                self._grad_scaler.step(opt)
                self._grad_scaler.update()
            else:
                opt.step()
        if lr_scheduler is not None:
            lr_scheduler.step()
        opt.zero_grad()
        if self._ema_model is not None:
            self._ema_model.update(self._model)

    def fit_batch(self, batch, step=None, phase='train', model=None, opt=None, lr_scheduler=None):
        if model is None:
            model = self._model
        if opt is None:
            opt = self._opt
        if lr_scheduler is None:
            lr_scheduler = self._lr_scheduler
        if not self.cfg.use_tpu and self._is_cuda_available:
            pt_util.batch2device(batch, self._device)
        with torch.set_grad_enabled(phase == 'train'):
            if self.cfg.use_fp16:
                with torch.cuda.amp.autocast():
                    outputs = model(**batch, phase=phase)
            else:
                outputs = model(**batch, phase=phase)
            if phase != 'test':
                if self.cfg.use_fp16:
                    with torch.cuda.amp.autocast():
                        losses = self.calc_loss(batch, outputs, model=model)
                else:
                    losses = self.calc_loss(batch, outputs, model=model)
                loss = losses['loss']/self.cfg.accumulated_batch_size
            else:
                losses = None
            if phase == 'train':
                if self.cfg.use_fp16:
                    loss = self._grad_scaler.scale(loss)
                    loss.backward()
                elif self.cfg.use_deepspeed:
                    self._ds_model.backward(loss)
                else:
                    loss.backward()
                if step % self.cfg.accumulated_batch_size == 0:
                    self._step_opt(opt=opt, lr_scheduler=lr_scheduler)
        #if self.cfg.use_tpu and phase!='test':
        #    losses = xm.mesh_reduce('batch_loss_reduce', losses, pt_util.mesh_reduce_losses)
        return outputs, losses


    def tensor2numpy(self, tensor):
        if isinstance(tensor, list):
            return [self.tensor2numpy(t) for t in tensor]
        elif isinstance(tensor, np.ndarray):
            return tensor
        elif torch.is_tensor(tensor):
            return tensor.cpu().detach().numpy()
        else:
            return tensor

    def outputs2preds(self, outputs):
        preds = dict()
        for output in outputs:
            for k, v in output.items():
                if 'loss' not in k:
                    v = self.tensor2numpy(v)
                    if k not in preds:
                        preds[k] = []
                    if self.cfg.to_list:
                        preds[k].extend(list(v))
                    else:
                        preds[k].append(v)
        if self.cfg.do_concat:
            for k in preds:
                preds[k] = np.concatenate(preds[k],0)
        return preds

    @XLA
    def predict(self, ds, data_type='val', pt_model=None, **kwargs):
        if self._model is None:
            self.create_model(pt_model=pt_model)
        if self._ema_model is not None:
            model = self._ema_model.ema
            logger.info('use ema model to predict')
        else:
            model = self._model
        model.eval()
        if self.cfg.use_tpu:
            device = xm.xla_device()
            ds = pl.ParallelLoader(ds, [device]).per_device_loader(device)
        outputs = []
        if not isinstance(ds, tqdm):
            ds = tqdm(enumerate(ds), total=len(ds), desc="predict")
        for i, batch in ds:
            output, _ = self.run_batch(batch, phase='test', model=model, **kwargs)
            outputs.append(output)
        preds = self.outputs2preds(outputs)
        self.info('predict done')
        return preds

    def fit_epoch(self, ds, epoch, step, device=None, **kwargs):
        """

        :param ds:
        :param epoch:
        :param step: global step accross epochs
        :param opt:
        :param history:
        :param barlog:
        :param kwargs:
        :return:
        """
        loss_hist = LossHist()
        self._model.train()
        desc = 'Train Epoch:{}, lr:{:.8f}, loss:{}'
        miniters, verbose, is_master = self.cfg.verbose, self.cfg.verbose, True
        if hasattr(ds, 'sampler') and hasattr(ds.sampler, 'set_epoch'):
            ds.sampler.set_epoch(epoch)
        if self.cfg.use_tpu:
            device = xm.xla_device()
            ds = pl.ParallelLoader(ds, [device]).per_device_loader(device)

            if not xm.is_master_ordinal():
                is_master = False
                miniters = int(1e10)
                verbose = int(1e10)
                ds_itr = enumerate(ds)
            else:
                ds_itr = tqdm(enumerate(ds), total=len(ds), disable=verbose==0, miniters=miniters, desc=desc.format(epoch, self._opt.state_dict()['param_groups'][0]['lr'], ''))
        else:
            device = next(self._model.parameters()).device
            ds_itr = tqdm(enumerate(ds), total=len(ds), disable=verbose==0, miniters=miniters, desc=desc.format(epoch, self._opt.state_dict()['param_groups'][0]['lr'], ''))
        for i, batch in ds_itr:
            # do not know why miniters was reset to 1
            if isinstance(ds_itr, tqdm):
                ds_itr.miniters = miniters
            step += 1
            outputs, losses = self.fit_batch(batch, step)
            for k, v in losses.items():
                losses[k] = v.item()
            loss_hist.append(losses)
            #ds_itr.set_description()
            if step%verbose == 0 and is_master:
                ds_itr.set_description(desc.format(epoch, self._opt.state_dict()['param_groups'][0]['lr'], loss_hist.avg_output()))

            if step % (self.cfg.n_save_step) == 0:
                logger.info('%%%% save model to %s for step:%s', self.cfg.output_dir, step)
                self.save(global_step=step, save_opt=self.cfg.save_opt)

            if (i + 1) >= self.cfg.n_epoch_step:
                logger.info('max %s step per epoch reached', self.cfg.n_epoch_step)
                break
            if step % self.cfg.n_train_step == 0:
                break
        #logger.info('epoch:{}, '.format(epoch) + loss_hist.avg_output())
        losses = loss_hist.get_avg()
        if self.cfg.use_tpu:
            losses = xm.mesh_reduce('epoch loss reduce', losses, pt_util.mesh_reduce_losses)
        return step, losses

    def run_batch(self, batch, phase='val', model=None, **kwargs):
        outputs, losses = self.fit_batch(batch, phase=phase, model=model)
        new_outputs = dict()
        for k in self.cfg.batch_keys:
            if k in outputs:
                new_outputs[k] = outputs[k]
            elif k in batch:
                new_outputs[k] = batch[k]
        outputs = new_outputs
        if self.cfg.use_tpu and kwargs.get('xla_procs', self.cfg.xla_procs)>1:
            outputs = xm.mesh_reduce('run batch outputs reduce', outputs, pt_util.mesh_reduce_outputs)
        return outputs, losses

    def val_epoch(self, ds, epoch, model=None):
        loss_hist = LossHist()
        if model is None:
            model = self._model
            if self._ema_model is not None:
                model = self._ema_model.ema
                logger.info('use ema model for val')
        model.eval()
        outputs = []
        miniters, verbose, is_master = self.cfg.verbose, self.cfg.verbose, True
        desc = 'Val Epoch:{}, loss:{}'
        if self.cfg.use_tpu:
            device = xm.xla_device()
            ds = pl.ParallelLoader(ds, [device]).per_device_loader(device)
            if not xm.is_master_ordinal():
                miniters, verbose, is_master = int(1e10), int(1e10), False
                ds_itr = enumerate(ds)
            else:
                ds_itr = tqdm(enumerate(ds), disable=verbose==0, total=len(ds), miniters=miniters, desc=desc.format(epoch, ''))
        else:
            device = next(self._model.parameters()).device
            ds_itr = tqdm(enumerate(ds), total=len(ds), disable=verbose==0, miniters=miniters, desc=desc.format(epoch, ''))
        for i, batch in ds_itr:
            if isinstance(ds_itr, tqdm):
                ds_itr.miniters = miniters
            output, losses = self.run_batch(batch, phase='val', model=model)
            outputs.append(output)
            val_losses = OrderedDict([('val_' + key, v.item()) for key, v in losses.items()])
            loss_hist.append(val_losses)
            if (i+1)%verbose == 0 and is_master:
                ds_itr.set_description(desc.format(epoch, loss_hist.avg_output()))

        if self.cfg.predicting or self.cfg.scoring:
            preds = self.outputs2preds(outputs)
        else:
            preds = None

        losses = loss_hist.get_avg()
        if self.cfg.use_tpu:
            losses = xm.mesh_reduce('batch_loss_reduce', losses, pt_util.mesh_reduce_losses)
        return i+1, preds, losses

    def _gen_distribute_dl(self, dl, data_type):
        shuffle, drop_last, collate_fn = False, False, None
        if data_type=='train':
            shuffle, drop_last = True, True
        # use original configure
        collate_fn = getattr(dl, 'collate_fn', collate_fn)
        shuffle = getattr(dl, 'shuffle', shuffle)
        ds, drop_last = dl.dataset, dl.drop_last
        if isinstance(ds, torch.utils.data.IterableDataset):
            return dl
        else:
            sampler = torch.utils.data.distributed.DistributedSampler(ds, num_replicas=xm.xrt_world_size(), rank=xm.get_ordinal(), shuffle=shuffle)
            dl = torch.utils.data.DataLoader(ds, batch_size=self.cfg.batch_size//xm.xrt_world_size(), sampler=sampler, num_workers=self.cfg.n_dl_worker, drop_last=drop_last, collate_fn=collate_fn)
        return dl

    def info(self, text, *args):
        if self.cfg.use_tpu:
            if xm.is_master_ordinal():
                logger.info(text, *args)
        else:
            logger.info(text, *args)

    def gen_bn_loader(self, dl, batch_size):
        shuffle, drop_last, collate_fn = True, True, None
        collate_fn = getattr(dl, 'collate_fn', collate_fn)
        ds, drop_last = dl.dataset, dl.drop_last
        dl = torch.utils.data.DataLoader(ds, batch_size=self.cfg.swa_batch_size, num_workers=self.cfg.n_dl_worker, drop_last=drop_last, collate_fn=collate_fn)
        return dl

    def bn_update(self, loader, batch_size=None, pt_model=None, device=None):
        r"""Updates BatchNorm running_mean, running_var buffers in the model.
        It performs one pass over data in `loader` to estimate the activation
        statistics for BatchNorm layers in the model.
        Args:
            loader (torch.utils.data.DataLoader): dataset loader to compute the
                activation statistics on. Each data batch should be either a
                tensor, or a list/tuple whose first element is a tensor
                containing data.
            model (torch.nn.Module): model for which we seek to update BatchNorm
                statistics.
            device (torch.device, optional): If set, data will be trasferred to
                :attr:`device` before being passed into :attr:`model`.
        """
        logger.info('update batch normalization')
        if self._model is None:
            self.create_model(pt_model=pt_model)
        model = self._model
        if batch_size is not None and batch_size != loader.batch_size:
            loader = self.gen_bn_loader(loader, batch_size)

        if not _check_bn(model):
            logger.info('check bn failed, not update bn')
            return
        was_training = model.training
        model.train()
        momenta = {}
        model.apply(_reset_bn)
        model.apply(lambda module: _get_momenta(module, momenta))
        n = 0
        for i, input in enumerate(loader):
            if isinstance(input, dict):
                for k, v in input.items():
                    b = v.size(0)
                    break
            else:
                raise NotImplementedError

            momentum = b / float(n + b)
            for module in momenta.keys():
                module.momentum = momentum

            self.fit_batch(input, i, 'test')
            n += b

        model.apply(lambda module: _set_momenta(module, momenta))
        model.train(was_training)

    def swap_swa_sgd(self, ds, model):
        logger.info('swap swa weights')
        self._opt.swap_swa_sgd()
        if not self.cfg.swa_disable_bn:
            self.bn_update(ds, self.cfg.swa_batch_size)

    def get_start_epoch(self):
        start_epoch = 0
        if self._restored_epoch is not None:
            start_epoch = self._restored_epoch + 1
        return start_epoch

    @XLA
    def fit(self, train_ds=None, val_ds=None, opt=None, pt_model=None, **kwargs):
        self.info('start fit')
        if self._model is None:
            self.create_model(opt=opt, pt_model=pt_model, **kwargs)

        best_loss = np.inf;
        best_epoch, step, save_opt = -1, 0, self.cfg.save_opt
        start_epoch = self.get_start_epoch()
        for epoch in range(start_epoch, start_epoch + self.cfg.epochs):
            ## need this because numpy will use same random state for epoch fork subprocess
            np.random.seed(self.cfg.seed + epoch)
            random.seed(self.cfg.seed+epoch)

            self.info(os.linesep)
            loss = None
            if train_ds is not None and not self.cfg.only_validate:
                step, losses = self.fit_epoch(train_ds, epoch, step, **kwargs)
                loss = losses['loss']
                if (epoch+1)>=self.cfg.swa_start_epoch and self.cfg.swa_start_epoch>0:
                    if (epoch+1)==self.cfg.swa_start_epoch:
                        logger.info('start update swa')
                    self._opt.update_swa()
            preds = None
            if val_ds is not None and (epoch+1)>self.cfg.n_init_epoch and not self.cfg.no_validate:
                val_step, preds, val_losses = self.val_epoch(val_ds, epoch)
                val_loss = val_losses['val_loss']
                self.info('loss for epoch %s:, train: %s, val: %s', epoch,  loss, val_loss)
                if self.cfg.scoring:
                    s = self.score(val_ds, preds=preds)
                    if self.cfg.use_tpu:
                        s = xm.mesh_reduce('score reduce', s, np.mean)
                        self.info("reduced score is:%s", s)
                if val_loss < (best_loss*(1-self.cfg.es_min_delta)):
                    best_loss = val_loss
                    best_epoch = epoch
                    if self.cfg.save_best and (epoch+1)>self.cfg.n_init_epoch:
                        if self.cfg.swa_start_epoch>0:
                            self.swap_swa_sgd(train_ds, self._model)
                        self.save(save_opt=save_opt, epoch=epoch)
                if self._should_stop(best_loss, val_loss, best_epoch, epoch):
                    self.info(
                        "best_loss:%s, best_epoch:%s, current_epoch:%s, val_loss:%s, es_min_delta:%s, without improvement for %s epochs, train done",
                        best_loss, best_epoch, epoch, val_loss, self.cfg.es_min_delta, self.cfg.n_es_epoch)
                    save_opt = False
                    break
                elif self.cfg.only_validate:
                    break
                else:
                    self.info('best val loss:%s, best epoch:%s', best_loss, best_epoch)
            if ((epoch + 1) % (self.cfg.n_save_epoch) == 0) and (epoch+1)>self.cfg.n_init_epoch:
                if self.cfg.swa_start_epoch > 0:
                    self.swap_swa_sgd(train_ds, self._model)
                self.info('%%%% save model to %s for epoch:%s', self.cfg.output_dir, epoch)
                self.save(global_step=step, save_opt=save_opt, epoch=epoch)
            if not self.cfg.only_validate and step % self.cfg.n_train_step == 0:
                self.info('total train step %s done', self.cfg.n_train_step)
                break
        if self.cfg.save or self.cfg.save_best or (self.cfg.n_save_epoch < self.cfg.N_INF) or (self.cfg.n_save_step < self.cfg.N_INF):
            if not self.cfg.only_validate and not self.cfg.save_best:
                if self.cfg.swa_start_epoch > 0:
                    self.swap_swa_sgd(train_ds, self._model)
                self.save(save_opt=save_opt, epoch=epoch)
            if preds is not None and self.cfg.save_pred:
                self.save_predict(preds, '_val')

    def save_predict(self, pred, suffix=''):
        if not self.cfg.use_tpu:
            super(PTModel, self).save_predict(pred, suffix=suffix)
        else:
            if xm.is_master_ordinal():
                super(PTModel, self).save_predict(pred, suffix=suffix)

    def score(self, ds, preds=None):
        raise NotImplemented


class ModelEma:
    """ Model Exponential Moving Average
    ref:https://github.com/rwightman/efficientdet-pytorch
    Keep a moving average of everything in the model state_dict (parameters and buffers).

    This is intended to allow functionality like
    https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage

    A smoothed version of the weights is necessary for some training schemes to perform well.
    E.g. Google's hyper-params for training MNASNet, MobileNet-V3, EfficientNet, etc that use
    RMSprop with a short 2.4-3 epoch decay period and slow LR decay rate of .96-.99 requires EMA
    smoothing of weights to match results. Pay attention to the decay constant you are using
    relative to your update count per epoch.

    To keep EMA from using GPU resources, set device='cpu'. This will save a bit of memory but
    disable validation of the EMA weights. Validation will have to be done manually in a separate
    process, or after the training stops converging.

    This class is sensitive where it is initialized in the sequence of model init,
    GPU assignment and distributed training wrappers.
    I've tested with the sequence in my own train.py for torch.DataParallel, apex.DDP, and single-GPU.
    """
    def __init__(self, model, decay=0.9999, device='', resume=''):
        # make a copy of the model for accumulating moving average of weights
        self.ema = deepcopy(model)
        self.ema.eval()
        self.decay = decay
        self.device = device  # perform ema on different device from model if set
        if device:
            self.ema.to(device=device)
        self.ema_has_module = hasattr(self.ema, 'module')
        if resume:
            self._load_checkpoint(resume)
        for p in self.ema.parameters():
            p.requires_grad_(False)

    def _load_checkpoint(self, checkpoint_path):
        checkpoint = torch.load(checkpoint_path, map_location='cpu')
        assert isinstance(checkpoint, dict)
        if 'state_dict_ema' in checkpoint:
            new_state_dict = OrderedDict()
            for k, v in checkpoint['state_dict_ema'].items():
                # ema model may have been wrapped by DataParallel, and need module prefix
                if self.ema_has_module:
                    name = 'module.' + k if not k.startswith('module') else k
                else:
                    name = k
                new_state_dict[name] = v
            self.ema.load_state_dict(new_state_dict)
            logger.info("Loaded state_dict_ema")
        else:
            logger.warning("Failed to find state_dict_ema, starting from loaded model weights")

    def update(self, model):
        # correct a mismatch in state dict keys
        needs_module = hasattr(model, 'module') and not self.ema_has_module
        with torch.no_grad():
            msd = model.state_dict()
            for k, ema_v in self.ema.state_dict().items():
                if needs_module:
                    k = 'module.' + k
                model_v = msd[k].detach()
                if self.device:
                    model_v = model_v.to(device=self.device)
                ema_v.copy_(ema_v * self.decay + (1. - self.decay) * model_v)




import logging
from tensorflow.keras.layers import Conv2D, MaxPool2D, Dropout, Cropping2D, Conv2DTranspose, UpSampling2D
import tensorflow as tf

from mautil.tf.nn import NN

logger = logging.getLogger(__name__)

def calc_cropping_2d(shape, target_shape):
    dh = shape[0] - target_shape[0]
    dw = shape[1] - target_shape[1]
    crop_top, crop_left = dh//2,  dw//2
    crop_bottom, crop_right = dh-crop_top, dw-crop_left
    return crop_top, crop_bottom, crop_left, crop_right


class DownBlock(NN):
    def __init__(self, name, filters, kernel_size, width=2, strides=(1, 1), dropout=0.1, padding='same', activation='relu',
                 use_batch_norm=True, **kwargs):
        super(DownBlock, self).__init__(name=name)
        self.filters = filters
        self.kernel_size = kernel_size
        self.width = width
        self.strides = strides
        self.dropout = dropout
        self.padding=padding
        self.activation = activation
        self.use_batch_norm = use_batch_norm
        self.kwargs = kwargs

        self.block_layers = self.create_layers()

    def create_layers(self):
        layers = []; strides = self.strides
        for i in range(self.width):
            if i!=0:
                strides = (1,1)
            layers.append(Conv2D(self.filters, self.kernel_size, padding=self.padding, strides=strides, **self.kwargs))
            if self.use_batch_norm:
                layers.append(tf.keras.layers.BatchNormalization())
            if self.dropout>0:
                logger.info('%s use dropout %s', self.name, self.dropout)
                layers.append(Dropout(self.dropout))
            layers.append(tf.keras.layers.Activation(self.activation))
        return layers

    def call(self, inputs, training=None, mask=None):
        for layer in self.block_layers:
            inputs = layer(inputs, training=training)
        return inputs


class UpBlock(DownBlock):
    def create_layers(self):
        layers = []
        for i in range(self.width+1):
            layer = tf.keras.Sequential()
            if i==0:
                layer.add( Conv2DTranspose(self.filters, self.kernel_size, strides=self.strides, padding=self.padding,
                                    **self.kwargs))
            else:
                layer.add(Conv2D(self.filters, self.kernel_size, padding=self.padding, **self.kwargs))
            if self.use_batch_norm:
                layer.add(tf.keras.layers.BatchNormalization())
            if self.dropout > 0:
                logger.info('%s use dropout %s', self.name, self.dropout)
                layer.add(Dropout(self.dropout))
            layer.add(tf.keras.layers.Activation(self.activation))
            layers.append(layer)
        return layers

    def call(self, inputs, skip, training=None, mask=None):
        inputs = self.block_layers[0](inputs, training=training)
        if self.padding == 'same':
            crop_top, crop_bottom, crop_left, crop_right = calc_cropping_2d(inputs.shape[1:], skip.shape[1:])
            if crop_top!=0 or crop_bottom != 0 or crop_left !=0 or crop_right !=0:
            #if 1==1:
                logger.info('cropped inputs shape %s to skip shape %s', tf.shape(inputs), tf.shape(skip))
                inputs = Cropping2D(cropping=((crop_top, crop_bottom), (crop_left, crop_right)))(inputs, training=training)
        elif self.padding =='valid':
            crop_top, crop_bottom, crop_left, crop_right = calc_cropping_2d(skip.shape[1:], inputs.shape[1:])
            skip = Cropping2D(cropping=((crop_top, crop_bottom), (crop_left, crop_right)))(skip, training=training)

        inputs = tf.concat([skip, inputs], -1)
        for layer in self.block_layers[1:]:
            inputs = layer(inputs, training=training)
        return inputs


class Stack(NN):
    def __init__(self, name, blocks, **kwargs):
        super(Stack, self).__init__(name=name, **kwargs)
        self.blocks = blocks

    def call(self, inputs, skips=None, training=None, mask=None):
        outputs = []
        for i, block in enumerate(self.blocks):
            if skips is not None:
                inputs = block(inputs, skip=skips[i], training=training, mask=mask)
            else:
                inputs = block(inputs, training=training, mask=mask)
            outputs.append(inputs)
        return outputs


class Unet(NN):
    def __init__(self, name='unet', filters=64, kernel_size=3, depth=5, width=2, up_dropout=0.1, strides=(2,2), padding='same',
                 activation='relu', down_cls=DownBlock, up_cls=UpBlock, down_dropout=0.1, **kwargs):
        """
        :param name:
        :param filters:
        :param kernel_size:
        :param depth: num of downsample plus one
        :param width: num of conv layer exclude up/down sample layer
        :param dropout:
        :param pool_size:
        :param padding:
        :param activation:
        :param kwargs:
        """

        super(Unet, self).__init__(name=name)
        self.return_last = kwargs.pop('return_last', False)
        self.filters = []
        if not isinstance(filters, list):
            for i in range(depth):
                self.filters.append(filters*(2**i))
        else:
            self.filters = filters

        self.up_dropout = up_dropout
        self.down_dropout = down_dropout
        if not isinstance(up_dropout, list):
            self.up_dropout = [up_dropout for i in range(depth)]
        if not isinstance(down_dropout, list):
            self.down_dropout = [down_dropout for i in range(depth)]
        self.strides = strides
        if not isinstance(strides, list):
            self.strides = [tuple(strides) for i in range(depth)]



        down_blocks = []
        name = 'down_block_{}'
        for i in range(depth):
            strides= self.strides[i]
            use_batch_norm = False if i==0 else True
            down_blocks.append(down_cls(name.format(i), self.filters[i], kernel_size, width, strides, self.down_dropout[i],
                                        padding, activation, use_batch_norm=use_batch_norm, **kwargs))
        self.down_stack = Stack('down_stack', down_blocks)

        up_blocks = []
        name = 'up_block_{}'
        for i in range(depth-2, -1, -1):
            up_blocks.append(up_cls(name.format(i), self.filters[i], kernel_size, width, self.strides[i+1], self.up_dropout[i],
                                    padding, activation, **kwargs))
        self.up_stack = Stack('up_stack', up_blocks)

    def call(self, inputs, training=None, mask=None):
        down_outputs = self.down_stack(inputs, training=training, mask=mask)
        x = down_outputs[-1]
        skips = down_outputs[:-1][::-1]
        up_outputs = self.up_stack(x, skips, training=training, mask=mask)
        outputs = [x] + up_outputs
        if self.return_last:
            outputs = outputs[-1]
        return outputs


if __name__ == "__main__":
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s',
                        datefmt='%m/%d/%Y %H:%M:%S', level=logging.DEBUG)
    import numpy as np
    depth = 3
    filters = 2
    kernel_size = 3
    strides = [(1,1), (2,2), (2,2)]
    inputs = np.random.rand(2, 48, 51, 3).astype(np.float32)

    unet = Unet(filters=filters, kernel_size=kernel_size, depth=depth, strides=strides, padding='same')
    outputs = unet(inputs)
    assert outputs[-1].shape == list(inputs.shape[:3]) + [filters], outputs[-1].shape

    unet = Unet(filters=filters, kernel_size=kernel_size, depth=depth, strides=strides, padding='valid')
    outputs = unet(inputs)
    assert outputs[-1].shape == (2, 8, 8, filters), outputs[-1].shape

    print('^_^')






import tensorflow as tf


class BiRNN(tf.keras.layers.Layer):
    def __init__(self, dim, return_sequences=True, rnn=tf.keras.layers.GRU, **kwargs):
        super(BiRNN, self).__init__(**kwargs)
        self._rnn_fw = rnn(dim, return_sequences=return_sequences, go_backwards=False, **kwargs)
        self._rnn_bw = rnn(dim, return_sequences=return_sequences, go_backwards=False, **kwargs)

    def call(self, seqs, seqs_len, seq_axis=1, batch_axis=0):
        fw_seqs = seqs
        bw_seqs = tf.reverse_sequence(seqs, seqs_len, seq_axis=seq_axis, batch_axis=batch_axis)
        fw_out = self._rnn_fw(fw_seqs)
        bw_out = self._rnn_bw(bw_seqs)
        bw_out = tf.reverse_sequence(bw_out, seqs_len, seq_axis=seq_axis, batch_axis=batch_axis)
        out = tf.concat([fw_out, bw_out], -1)
        return out


import logging, os, json
import random
from collections import OrderedDict, defaultdict
import numpy as np

import tensorflow as tf
from tensorflow.python.keras.callbacks import CallbackList
from tensorflow.keras.mixed_precision import experimental as mixed_precision
import tensorflow_addons as tfa
from tensorflow_addons.optimizers import MovingAverage

import mautil as mu
from mautil.basic_models import CFG as BasicCFG, Model as BasicModel
from mautil.tf import util as tf_util
from .optimizer import WarmupWrap


logger = logging.getLogger(__name__)


class CFG(BasicCFG):
    def __init__(self):
        super(CFG, self).__init__()
        self.optimizer = 'adam'
        self.loss = 'mse'
        self.global_step = None
        self.gradient_clip = None
        self.dropout = None
        self.keep_checkpoint_every_n_hours = 1000000000
        self.save_keep = 1000000000
        self.save_summary_step = 1000000000
        self.use_tpu = False
        self.use_tffit = True
        self.ema = 0
        self.tpu_name = None
        self.warm_start_path = None  # used by tpu
        self.tpu_zone = None
        self.gcp_project = None
        self.num_core_per_host = 1
        self.num_hosts = 1
        self.tpu_loop_iterations = 2
        self.run_eagerly = False
        self.accumulated_batch_size = 1
        self.compile_model = False

    def dump(self, fpath):
        tf.io.write_file(fpath, json.dumps(self.__dict__))

    def load(self, fpath):
        cfg = json.loads(tf.io.read_file(fpath).numpy())
        self.update(cfg)

class ModelCheckpoint(tf.keras.callbacks.ModelCheckpoint):
    def __init__(self, filepath, monitor='val_loss', verbose=1, save_best_only=False, save_weights_only=True, mode='auto',
                save_freq=int(1e16), **kwargs):
        self._save_on_train_end = kwargs.pop('save_on_train_end', False)
        super(ModelCheckpoint, self).__init__(filepath, monitor=monitor, verbose=verbose, save_best_only=save_best_only,
                                              save_weights_only=save_weights_only, mode=mode, save_freq=int(1e32), **kwargs)

    def _save_model(self, epoch, logs):
        if isinstance(self.model.opt, MovingAverage):
            self.model.opt.assign_average_vars(self.model.variables)
        super(ModelCheckpoint, self)._save_model(epoch, logs)

    def on_train_end(self, logs=None):
        if self._save_on_train_end:
            self._save_model(epoch=self._current_epoch, logs=logs)
        super(ModelCheckpoint, self).on_train_end(logs)


class TFModel(BasicModel):
    cfg = CFG()
    def __init__(self, name='TFModel', cfg={}):
        super(TFModel, self).__init__(name, cfg)
        self._input_features = self.get_input_features()
        self._called = False  # if run eagerly need call once to create weights
        self._restored = False
        self._accumulated_grads = None
        self._batch_id = None
        self._opt = None
        self._trainable_variables = None
        self._run_batch_graph_func = None
        self._fit_batch_graph_func = None
        self._val_batch_graph_func = None
        self._fit_func = None
        self.strategy = None

    def get_input_features(self):
        """

        :return: OrderedDict with InputFeature instance. For targets and smaple_weights, must named endswith targets,
        smaple_weights
        """
        input_features = OrderedDict()
        return input_features

    def get_keras_batch_inputs_and_types(self):
        inputs, inputs_types = OrderedDict(), OrderedDict()
        for name, fea in self._input_features.items():
            inputs[fea.name] = None
            inputs_types[fea.name] = fea.dtype
        return inputs, inputs_types

    def create_keras_inputs(self, input_features):
        inputs = OrderedDict()
        for k, fea in input_features.items():
            inputs[fea.name] = tf.keras.Input(tuple(fea.shape), dtype=fea.dtype, name=fea.name)
        return inputs

    def main_nn(self, inputs):
        outputs = OrderedDict()
        outputs['dummy'] = inputs['dummy']
        return outputs

    def _create_keras_model(self, inputs, outputs):
        model = tf.keras.Model(inputs=tuple(inputs.values()), outputs=outputs)
        pred_model = model
        return model, pred_model

    def create_keras_model(self, inputs, compile=True):
        outputs = self.main_nn(inputs)
        model, pred_model = self._create_keras_model(inputs, outputs)
        if compile:
            self.compile_model(model)
        return model, pred_model

    def compile_model(self, model):
        optimizer = self._create_optimizer(self.cfg.opt, lr=self.cfg.lr, n_lr_decay_step=self.cfg.n_lr_decay_step,
                                    lr_decay_rate=self.cfg.lr_decay_rate, n_lr_warmup_step=self.cfg.n_lr_warmup_step,
                                          ema=self.cfg.ema, weight_decay=self.cfg.weight_decay
                                    )
        loss = self.get_keras_loss()
        model.compile(optimizer=optimizer, loss=loss, run_eagerly=self.cfg.run_eagerly)

    def create_model(self):
        tf.random.set_seed(self.cfg.seed)
        np.random.seed(self.cfg.seed)
        random.seed(self.cfg.seed)
        if self.cfg.use_fp16:
            tf.config.optimizer.set_jit(True)
            tf.config.optimizer.set_experimental_options(
                {"auto_mixed_precision": True})
            #policy = mixed_precision.Policy('mixed_float16')
            #mixed_precision.set_policy(policy)


        if self.cfg.use_tpu:
            resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu=self.cfg.tpu_name, zone=self.cfg.tpu_zone,
                                                                         project=self.cfg.gcp_project)
            tf.config.experimental_connect_to_cluster(resolver)
            tf.tpu.experimental.initialize_tpu_system(resolver)
            self.strategy = tf.distribute.experimental.TPUStrategy(resolver)
            with self.strategy.scope():
                self.inputs = self.create_keras_inputs(self._input_features)
                self._model, self._pred_model = self.create_keras_model(self.inputs, compile=self.cfg.compile_model)
                self._opt = self._create_optimizer(self.cfg.opt, lr=self.cfg.lr, n_lr_decay_step=self.cfg.n_lr_decay_step,
                                                   lr_decay_rate=self.cfg.lr_decay_rate, n_lr_warmup_step=self.cfg.n_lr_warmup_step,
                                                   ema=self.cfg.ema, weight_decay=self.cfg.weight_decay
                                                   )
                if self.cfg.restore:
                    self.restore(epoch=self.cfg.restore_epoch)
        else:
            self.inputs = self.create_keras_inputs(self._input_features)
            self._model, self._pred_model = self.create_keras_model(self.inputs, compile=self.cfg.compile_model)
            if self.cfg.restore:
                self.restore(epoch=self.cfg.restore_epoch)

        if self.cfg.recompute:
            self._fit_func = tf.recompute_grad(self._model)
        else:
            self._fit_func = self._model


    def get_keras_loss(self):
        """

        :return: keras loss then compile the model
        """
        return self.cfg.loss

    def create_optimizer(self, optimizer_name, lr, wd, **kwargs):
        if optimizer_name == 'adam':
            opt = tf.keras.optimizers.Adam(learning_rate=lr, **kwargs)
        elif optimizer_name == 'sgd':
            opt = tf.keras.optimizers.SGD(learning_rate=lr, **kwargs)
        elif optimizer_name == 'adamW':
            ## tempoary solution
            assert wd is not None
            if isinstance(wd, float):
                opt = tfa.optimizers.AdamW(learning_rate=lr, weight_decay=wd, **kwargs)
            else:
                opt = tfa.optimizers.AdamW(learning_rate=lr, weight_decay=lambda:None, **kwargs)
                opt.weight_decay = lambda:wd(opt.iterations)
        else:
            raise Exception('unknown optmizer:{}'.format(optimizer_name))

        return opt

    def _create_optimizer(self, optimizer_name, lr, **kwargs):
        n_lr_warmup_step = kwargs.pop('n_lr_warmup_step', 0)
        n_lr_decay_step = kwargs.pop('n_lr_decay_step')
        lr_decay_rate = kwargs.pop('lr_decay_rate', 1)
        weight_decay = kwargs.pop('weight_decay', 0)
        wd = None
        if weight_decay>0:
            wd = weight_decay
        ema = kwargs.pop('ema', 0)

        if lr_decay_rate <1:
            lr = tf.keras.optimizers.schedules.ExponentialDecay(lr, n_lr_decay_step, lr_decay_rate)
            if wd is not None:
                wd = tf.keras.optimizers.schedules.ExponentialDecay(weight_decay, n_lr_decay_step, lr_decay_rate)
        if n_lr_warmup_step >0:
            lr = WarmupWrap(lr, n_lr_warmup_step)
            if wd is not None:
                wd = WarmupWrap(wd, n_lr_warmup_step)

        opt = self.create_optimizer(optimizer_name, lr, wd=wd, **kwargs)

        if ema > 0:
            opt = MovingAverage(opt, average_decay=ema)
        if self.cfg.use_fp16:
            opt = mixed_precision.LossScaleOptimizer(opt, loss_scale='dynamic')
        return opt

    def save(self, epoch=None, save_path=None):
        super(TFModel, self).save()
        if save_path is None:
            save_name = 'model-{epoch:04d}.ckpt'.format(epoch=epoch) if epoch else 'model.ckpt'
            save_path = self.gen_fname(save_name)

        self._model.save_weights(save_path)
        logger.info("Model saved to file:{}".format(save_path))

    def get_checkpoint_path(self, model_dir=None, epoch=None):
        if epoch is None:
            ckpt = tf.train.get_checkpoint_state(model_dir)
            model_path = ckpt.model_checkpoint_path
        else:
            save_name = 'model-{epoch:04d}.ckpt'.format(epoch=epoch)
            model_path = os.path.join(model_dir, save_name)
        return model_path

    def _get_save_vars(self):
        return tf.trainable_variables()

    def _get_restore_vars(self):
        return tf.trainable_variables()

    def _get_restore_assignment_map(self, var_list, model_path):
        (assignment_map, initialized_variable_names) = tf_util.get_assignment_map_from_checkpoint(var_list, model_path)
        assignment_map = {var.op.name: var for var in var_list if var.name in initialized_variable_names}

        return assignment_map, initialized_variable_names

    def restore(self, epoch=None, var_list=None, model_dir=None):
        # restore cfg first
        super(TFModel, self).restore()
        if self._model is None:
            self.create_model()

        if not self._restored:
            if model_dir is None:
                model_dir = self.gen_fname()
            model_path = self.get_checkpoint_path(model_dir, epoch)
            self._model.load_weights(model_path)
            self._restored = True
            logger.info('weights restored from %s', model_path)

    def get_modelcheckpoint_callback(self):
        model_path = self.gen_fname('model-{epoch:04d}.ckpt')
        cp = ModelCheckpoint(filepath=model_path, save_on_train_end=self.cfg.save, save_best_only=self.cfg.save_best)
        return cp

    def get_callbacks(self):
        callbacks = []
        if not self.cfg.no_validate:
            callbacks.append(tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=self.cfg.patience,
                                                              min_delta=self.cfg.es_min_delta,
                                                              ))
        callbacks.append(self.get_modelcheckpoint_callback())
        return callbacks

    def calc_loss(self, inputs, outputs):
        """

        :param outputs:
        :param targets:
        :return:  dictionary of losses
        """
        raise {}

    def _calc_loss(self, inputs, outputs):
        losses = self.calc_loss(inputs, outputs)
        if self.strategy is not None:
            for k, v in losses.items():
                if v.get_shape().ndims==0:
                    v = tf.stack([v])
                losses[k] = tf.nn.compute_average_loss(v, global_batch_size=self.strategy.num_replicas_in_sync)
        return losses

    def fit_batch(self, batch):
        with tf.GradientTape() as tape:
            outputs = self._fit_func(batch, training=tf.constant(True))
            losses = self._calc_loss(batch, outputs)
            loss = losses['loss']
            if self.cfg.use_fp16:
                loss = self._opt.get_scaled_loss(loss)
        vars = self._get_train_vars()
        grads = self.calc_grads(tape, vars, loss)
        self.apply_grads(vars, grads)
        return grads, losses

    @tf.function
    def fit_batch_graph(self, batch):
        return self.fit_batch(batch)

    def _fit_batch_distribute(self, inputs_with_signatures):
        inputs, input_signatures = inputs_with_signatures
        if self._fit_batch_graph_func is None:
            with self.strategy.scope():
                self._fit_batch_graph_func = tf.function(self.fit_batch, input_signature=(input_signatures,))
        grads, losses = self.strategy.experimental_run_v2(self._fit_batch_graph_func, args=(inputs,))
        losses = self._agg_distribute_loss(losses)
        return grads, losses

    def _fit_batch(self, inputs_with_signatures):
        if self.strategy is not None:
            return self._fit_batch_distribute(inputs_with_signatures)

        inputs, input_signatures = inputs_with_signatures
        if self.cfg.run_eagerly:
            return self.fit_batch(inputs)
        else:
            if self._fit_batch_graph_func is None:
                self._fit_batch_graph_func = self.fit_batch_graph.get_concrete_function(input_signatures)
            return self._fit_batch_graph_func(*inputs)


    def apply_grads(self, vars, grads):
        #vars = self.get_train_vars()

        if not self.cfg.accumulated_batch_size > 1:
            self._opt.apply_gradients(zip(grads, vars))

    def calc_grads(self, tape, vars, loss):
        grads = tape.gradient(loss, vars)
        if self.cfg.use_fp16:
            grads = self._opt.get_unscaled_gradients(grads)
        if self.cfg.gradient_clip is not None:
            (grads, gnorm) = tf.clip_by_global_norm(grads, clip_norm=self.cfg.gradient_clip)
        return grads

    def apply_accumulated_grads(self, vars, grads):
        if self._accumulated_grads is None:
            self._accumulated_grads = [tf_util.flat_gradients(grad)/self.cfg.accumulated_batch_size for grad in grads]
        else:
            for i, grad in enumerate(grads):
                self._accumulated_grads[i] += tf_util.flat_gradients(grad)/self.cfg.accumulated_batch_size
        if (self._batch_id + 1) % self.cfg.accumulated_batch_size == 0:
            grads_and_vars = zip(self._accumulated_grads, vars)
            self._opt.apply_gradients(grads_and_vars)
            self._accumulated_grads = None

    def apply_accumulated_grads_bak(self, vars, grads):
        if self._accumulated_grads is None:
            self._accumulated_grads = {tv.name:tf.Variable(tf.zeros_like(tv.read_value()), trainable=False) for tv in
                                       self._get_train_vars()}
        for j, (grad, var) in enumerate(zip(grads, vars)):
            if grad is not None:
                self._accumulated_grads[var.name].assign_add(grad/self.cfg.accumulated_batch_size)
        if (self._batch_id + 1) % self.cfg.accumulated_batch_size == 0:
            grads_and_vars = [(self._accumulated_grads[var.name], var) for var in vars]
            self._opt.apply_gradients(grads_and_vars)
            for j, (grad, var) in enumerate(zip(grads, vars)):
                if grad is not None:
                    self._accumulated_grads[var.name].assign(tf.zeros_like(grad))

    def get_train_vars(self):
        return self._model.trainable_variables

    def _get_train_vars(self):
        if self._trainable_variables is None:
            self._trainable_variables = self.get_train_vars()
        return self._trainable_variables

    def outputs2preds(self, outputs):
        preds = dict()
        for output in outputs:
            for k, v in output.items():
                if k not in preds:
                    preds[k] = []
                if self.cfg.to_list:
                    preds[k].extend(list(v))
                else:
                    preds[k].append(v)
        if self.cfg.do_concat:
            for k in preds:
                preds[k] = np.concatenate(preds[k],0)
        return preds

    def predict(self, ds, **kwargs):
        outputs = []
        for i, batch in enumerate(ds):
            output = self.predict_batch(batch, **kwargs)
            outputs.append(output)
        preds = self.outputs2preds(outputs)
        logger.info('predict done')
        return preds

    def process_batch_output(self, output, batch):
        for k in self.cfg.batch_keys:
            output[k] = batch[k]
        for k, v in output.items():
            _v = v.numpy()
            output[k] = _v
        return output

    def predict_batch(self, batch, **kwargs):
        inputs = self._get_model_inputs_tuple(batch)
        output = self._run_batch(inputs, **kwargs)
        output = self.process_batch_output(output, batch)
        return output

    def run_batch_to_init_weights(self, inputs):
        self._run_batch(inputs)
        self._called = True

    def _get_model_inputs_tuple(self, batch):
        inputs = [];
        input_signatures = []
        for name, fea in self._input_features.items():
            if name in batch:
                inputs.append(batch[name])
                input_signatures.append(tf.TensorSpec(shape=tuple([None] + fea.shape), dtype=fea.dtype))
        return tuple(inputs), tuple(input_signatures)

#    def _get_model_inputs_tuple(self, batch):
#        inputs = []; input_signatures = []
#        for name, fea in self._input_features.items():
#            if name in batch:
#                inputs.append(batch[name])
#                input_signatures.append(tf.TensorSpec(shape=None, dtype=fea.dtype))
#        return tuple(inputs), tuple(input_signatures)

    def fit_epoch(self, ds, epoch, step, opt, callbacks, **kwargs):
        """

        :param ds:
        :param epoch:
        :param step: global step accross epochs
        :param opt:
        :param history:
        :param barlog:
        :param kwargs:
        :return:
        """

        callbacks.on_epoch_begin(epoch)
        for i, batch in enumerate(ds):
            input_batch = self._get_model_inputs_tuple(batch)
            self._batch_id = i
            if not self._called:
                self.run_batch_to_init_weights(input_batch)
            step += 1
            callbacks.on_batch_begin(input_batch)
            grads, losses = self._fit_batch(input_batch)
            for k, v in losses.items():
                losses[k] = v.numpy()
            self._on_batch_end(input_batch, i, callbacks, losses)

            if self.cfg.accumulated_batch_size>1:
                self.apply_accumulated_grads(self._get_train_vars(), grads)

            if step % self.cfg.n_save_step == 0:
                logger.info('%%%% save model to %s for step:%s', self.cfg.output_dir, step)
                self.save(epoch=step)

            if (i + 1) >= self.cfg.n_epoch_step:
                logger.info('max %s step per epoch reached', self.cfg.n_epoch_step)
                break
            if step % self.cfg.n_train_step == 0:
                break
        #logger.info('%%%%train epoch:%s done, loss:%s', epoch, metric.result().numpy())
        callbacks.params['steps'] = i+1
        callbacks.set_params(callbacks.params)
        callbacks.on_epoch_end(epoch)
        return step

    def _on_batch_end(self, batch, i, callbacks, losses):
        if i == 0:
            callbacks.params['metrics'] = losses.keys()
            callbacks.set_params(callbacks.params)
        logs = OrderedDict()
        logs['size'] = 1
        logs.update(losses)
        callbacks.on_batch_end(batch, logs=logs)

    def _create_callbacks(self, steps=None, do_validation=False):
        callback_params = {
            'batch_size': self.cfg.batch_size,
            'samples': None,
            'steps': steps,
            'verbose': self.cfg.verbose,
            'epochs': self.cfg.epochs,
            'do_validation': True,
            'metrics': None,
        }
        history = tf.keras.callbacks.History()
        barlog = tf.keras.callbacks.ProgbarLogger('steps')
        baselogger = tf.keras.callbacks.BaseLogger()
        callbacks = CallbackList([baselogger, history, barlog])
        callbacks.set_params(callback_params)
        return callbacks, history

    @tf.function
    def run_batch_graph(self, inputs):
        return self.run_batch(inputs)

    def _run_batch_distribute(self, inputs_with_signatures):
        inputs, input_signatures = inputs_with_signatures
        if self._run_batch_graph_func is None:
            with self.strategy.scope():
                self._run_batch_graph_func = tf.function(self.run_batch, input_signature=(input_signatures,))
        outputs = self.strategy.experimental_run_v2(self._run_batch_graph_func, args=(inputs,))
        outputs = self._agg_distribute_outputs(outputs)
        return outputs

    def run_batch(self, inputs):
        return self._pred_model(inputs, training=False)

    def _run_batch(self, inputs_with_signatures, **kwargs):
        if self.strategy is not None:
            return self._run_batch_distribute(inputs_with_signatures)

        inputs, input_signatures = inputs_with_signatures
        if self.cfg.run_eagerly:
            outputs = self.run_batch(inputs)
        else:
            if self._run_batch_graph_func is None:
                self._run_batch_graph_func = self.run_batch_graph.get_concrete_function(input_signatures)
            outputs = self._run_batch_graph_func(*inputs)
        return outputs

    def val_batch(self, inputs):
        outputs = self._pred_model(inputs, training=tf.constant(False))
        losses = self._calc_loss(inputs, outputs)
        return losses, outputs

    @tf.function
    def val_batch_graph(self, inputs):
        return self.val_batch(inputs)

    @tf.function
    def _agg_distribute_loss(self, losses):
        new_losses = {}
        for k, v in losses.items():
            v = tf.stack(v.values, axis=0)
            v = tf.math.reduce_sum(v)
            new_losses[k] = v
        return new_losses

    def _agg_distribute_outputs(self, outputs):
        new_outputs = {}
        if isinstance(outputs, dict):
            for k, v in outputs.items():
                v = tf.concat(v.values, axis=0)
                new_outputs[k] = v
        elif isinstance(outputs, tuple):
            agg_outputs = []
            for output in outputs:
                agg_outputs.append(tf.concat(output.values, axis=0))
            new_outputs = tuple(agg_outputs)
        else:
            new_outputs = tf.concat(outputs.values, axis=0)
        return new_outputs

    def _val_batch_distribute(self, batch):
        inputs_with_signatures = self._get_model_inputs_tuple(batch)
        inputs, input_signatures = inputs_with_signatures
        if self._val_batch_graph_func is None:
            with self.strategy.scope():
                self._val_batch_graph_func = tf.function(self.val_batch, input_signature=(input_signatures,))
        losses, outputs = self.strategy.experimental_run_v2(self._val_batch_graph_func, args=(inputs,))
        losses = self._agg_distribute_loss(losses)
        outputs = self._agg_distribute_outputs(outputs)
        return losses, outputs

    def _val_batch(self, batch):
        if self.strategy is not None:
            return self._val_batch_distribute(batch)

        inputs, input_signatures = self._get_model_inputs_tuple(batch)
        if self.cfg.run_eagerly:
            return self.val_batch(inputs)
        else:
            if self._val_batch_graph_func is None:
                self._val_batch_graph_func = self.val_batch_graph.get_concrete_function(input_signatures)
            return self._val_batch_graph_func(*inputs)

    def val_epoch(self, ds, epoch, callbacks):
        callbacks.on_epoch_begin(epoch)
        outputs = []
        for i, batch in enumerate(ds):
            callbacks.on_batch_begin(batch)
            losses, output = self._val_batch(batch)
            if self.cfg.predicting or self.cfg.scoring:
                output = self.process_batch_output(output, batch)
            outputs.append(output)
            val_losses = OrderedDict([('val_' + key, v.numpy()) for key, v in losses.items()])
            self._on_batch_end(batch, i, callbacks, val_losses)

        if self.cfg.predicting or self.cfg.scoring:
            preds = self.outputs2preds(outputs)
        else:
            preds = None


        callbacks.params['steps'] = i+1
        callbacks.set_params(callbacks.params)
        callbacks.on_epoch_end(epoch)
        return i+1, preds
        #logger.info('%%%%val epoch:%s done, val loss:%s', epoch, history['metrics'])


    def tffit(self, train_ds=None, val_ds=None, opt=None, **kwargs):
        logger.info('start tffit')
        if self.strategy is not None:
            if train_ds is not None:
                train_ds = self.strategy.experimental_distribute_dataset(train_ds)
            if val_ds is not None:
                val_ds = self.strategy.experimental_distribute_dataset(val_ds)
        if self._model is None:
            self.create_model()
        if opt is not None:
            self._opt = opt
        elif self._opt is None:
            self._opt = self._create_optimizer(self.cfg.opt, lr=self.cfg.lr, n_lr_decay_step=self.cfg.n_lr_decay_step,
                                        lr_decay_rate=self.cfg.lr_decay_rate, n_lr_warmup_step=self.cfg.n_lr_warmup_step,
                                        ema=self.cfg.ema, weight_decay=self.cfg.weight_decay
                                        )
        steps_per_epoch = self.cfg.n_epoch_step if self.cfg.n_epoch_step != self.cfg.N_INF else None
        callbacks, history = self._create_callbacks(steps_per_epoch)
        val_callbacks, val_history = self._create_callbacks()
        callbacks.on_train_begin()
        val_callbacks.on_train_begin()

        best_loss = np.inf;
        best_epoch = -1;
        step = 0
        for epoch in range(self.cfg.epochs):
            print(os.linesep)
            if train_ds is not None and not self.cfg.only_validate:
                step = self.fit_epoch(train_ds, epoch, step, opt, callbacks, **kwargs)
                losses = history.history['loss']
                if len(losses)>1 and losses[-1]>(losses[-2]*(1+self.cfg.abnormal_loss_thr)):
                    logger.error("train loss increasing more than %s, abnormal", self.cfg.abnormal_loss_thr)
                    raise Exception("train loss increasing, abnormal")
                #history.on_epoch_end({'loss':loss})
            preds = None
            if val_ds is not None and (epoch+1)>self.cfg.n_init_epoch and not self.cfg.no_validate:
                val_step, preds = self.val_epoch(val_ds, epoch, val_callbacks)
                if self.cfg.scoring:
                    self.score(val_ds, preds)
                val_loss = val_history.history['val_loss'][-1]
                history.history.update(val_history.history)
                if val_loss < best_loss:
                    best_loss = val_loss
                    best_epoch = epoch
                if self._should_stop(best_loss, val_loss, best_epoch, epoch):
                    break
                elif self.cfg.only_validate:
                    break
                else:
                    logging.info('best val loss:%s, best epoch:%s', best_loss, best_epoch)
            self.cfg.history = history.history
            if self.cfg.verbose==0:
                history_output = dict()
                for k in history.history:
                    history_output[k] = history.history[k][-1]
                logging.info('loss for epoch %s: %s', epoch,  history_output)
            if not self.cfg.only_validate and step % self.cfg.n_train_step == 0:
                logger.info('total train step %s done', self.cfg.n_train_step)
                break
            if ((epoch + 1) % (self.cfg.n_save_epoch) == 0) and (epoch+1)>self.cfg.n_init_epoch:
                logger.info('%%%% save model to %s for epoch:%s', self.cfg.output_dir, epoch)
                self.save(epoch=epoch + 1)
        if isinstance(self._opt, MovingAverage) and not self.cfg.only_validate:
            self._opt.assign_average_vars(self._get_train_vars())
            logger.info('assigned ema vars')
        if self.cfg.save or (self.cfg.n_save_epoch < self.cfg.N_INF) or (self.cfg.n_save_step < self.cfg.N_INF):
            if not self.cfg.only_validate:
                self.save(epoch=epoch+1)
            if preds is not None:
                self.save_predict(preds, '_val')
        return history

    def fit(self, train_ds, val_ds=None, **kwargs):
        if self.cfg.use_tffit:
            return self.tffit(train_ds, val_ds, **kwargs)
        else:
            return self.kfit(train_ds, val_ds, **kwargs)

    def kfit(self, train_ds, val_ds=None, callbacks=None, **kwargs):
        if callbacks is None:
            callbacks = self.get_callbacks()
        if self._model is None:
            self.create_model()
        if self.cfg.n_epoch_step == self.cfg.N_INF:
            steps_per_epoch=None
        else:
            steps_per_epoch = self.cfg.n_epoch_step
        history = self._model.fit(train_ds, validation_data=val_ds, epochs=self.cfg.epochs, verbose=self.cfg.verbose, shuffle=False,
                                  validation_steps=self.cfg.validation_steps, validation_freq=self.cfg.validation_freq,
                                  initial_epoch=self.cfg.restore_epoch,
                                  steps_per_epoch=steps_per_epoch, callbacks=callbacks, **kwargs)
        return history

    def score(self, ds, preds=None):
        raise NotImplemented



import os, logging
from collections import OrderedDict
import tensorflow as tf

from mautil.tf.models import TFModel, CFG
from mautil.basic_models import InputFeature
import mautil as mu

logger = logging.getLogger(__name__)

class GAN(TFModel):
    """ A basic jointly training gan
    """
    cfg = TFModel.cfg.copy()
    cfg.noise_dim = 128
    cfg.sample_num = 16
    cfg.save_sample = False
    cfg.gen_weight = 1.0

    def __init__(self, name='GAN', cfg={}):
        super(GAN, self).__init__(name, cfg)
        self._samples = self.get_samples()

    def get_samples(self):
        samples = tf.random.normal([self.cfg.sample_num, self.cfg.noise_dim], seed=self.cfg.seed)
        return samples

    def get_input_features(self):
        input_features = super(GAN, self).get_input_features()
        input_features['img'] = InputFeature('img', [28, 28, 1], tf.float32)
        input_features['noise'] = InputFeature('noise', [self.cfg.noise_dim], tf.float32)
        #input_features['dis_real_targets'] = InputFeature('dis_real_targets', [], tf.float32)
        #input_features['dis_fake_targets'] = InputFeature('dis_fake_targets', [], tf.float32)
        #input_features['gen_targets'] = InputFeature('gen_targets', [], tf.float32)
        return input_features

    def main_nn(self, inputs):
        imgs = inputs['img']
        noise = inputs['noise']
        #noise = tf.random.normal([self.cfg.batch_size, self.cfg.noise_dim])
        gen_outputs = self.generator(noise)
        real_outputs = self.discrimator(imgs)
        fake_outputs = self.discrimator(gen_outputs)
        return fake_outputs, real_outputs, fake_outputs

    def create_keras_model(self, inputs):
        self.generator = self.create_generator()
        self.discrimator = self.create_discrimator()
        return super(GAN, self).create_keras_model(inputs, compile=False)

    def create_generator(self):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Dense(7 * 7 * 256, use_bias=False, input_shape=(self.cfg.noise_dim,)))
        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.LeakyReLU())

        model.add(tf.keras.layers.Reshape((7, 7, 256)))
        assert model.output_shape == (None, 7, 7, 256)

        model.add(tf.keras.layers.Conv2DTranspose(128, (5, 5), strides=(1, 1), padding='same', use_bias=False))
        assert model.output_shape == (None, 7, 7, 128)
        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.LeakyReLU())

        model.add(tf.keras.layers.Conv2DTranspose(64, (5, 5), strides=(2, 2), padding='same', use_bias=False))
        assert model.output_shape == (None, 14, 14, 64)
        model.add(tf.keras.layers.BatchNormalization())
        model.add(tf.keras.layers.LeakyReLU())

        model.add(tf.keras.layers.Conv2DTranspose(1, (5, 5), strides=(2, 2), padding='same', use_bias=False, activation='tanh'))
        assert model.output_shape == (None, 28, 28, 1)

        return model

    def create_discrimator(self):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Conv2D(64, (5, 5), strides=(2, 2), padding='same',
                                         input_shape=[28, 28, 1]))
        model.add(tf.keras.layers.LeakyReLU())
        model.add(tf.keras.layers.Dropout(0.3))

        model.add(tf.keras.layers.Conv2D(128, (5, 5), strides=(2, 2), padding='same'))
        model.add(tf.keras.layers.LeakyReLU())
        model.add(tf.keras.layers.Dropout(0.3))

        model.add(tf.keras.layers.Flatten())
        model.add(tf.keras.layers.Dense(1))

        return model

    def calc_loss(self, outputs, targets):
        gen_outputs, real_outputs, fake_outputs = outputs
        #gen_targets, real_targets, fake_targets = targets
        gen_targets, real_targets, fake_targets = tf.ones_like(gen_outputs), tf.ones_like(real_outputs), tf.zeros_like(fake_outputs)
        losses = OrderedDict()
        losses['gen_loss'] = self.cfg.gen_weight * tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(gen_targets, gen_outputs))
        real_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(real_targets, real_outputs))
        fake_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(fake_targets, fake_outputs))
        dis_loss = real_loss + fake_loss
        losses['real_loss'] = real_loss
        losses['fake_loss'] = fake_loss
        losses['dis_loss'] = dis_loss

        losses['loss'] = losses['gen_loss'] + losses['dis_loss']
        return losses

    def save_sample(self, epoch):
        fpath = self.gen_fname('sample_output', 'sample_{:04d}.dump'.format(epoch))
        gen_outputs = self.generator(self._samples)
        mu.dump(gen_outputs, fpath)
        logger.info('sample saved to %s', fpath)

    def load_sample(self, epochs=None):
        if epochs is None:
            epochs = self.cfg.epochs
        epoch_outputs = []
        for i in range(epochs):
            fpath = self.gen_fname('sample_output', 'sample_{:04d}.dump'.format(i))
            if os.path.exists(fpath):
                epoch_outputs.append(mu.load_dump(fpath))
            else:
                logger.error('sample outputs of epoch:%s does not exists:%s', i, fpath)
        return epoch_outputs

    def fit_epoch(self, ds, epoch, step, opt, callbacks, **kwargs):
        rst = super(GAN, self).fit_epoch(ds, epoch, step, opt, callbacks, **kwargs)
        if self.cfg.save_sample:
            self.save_sample(epoch)
        return rst

    def _get_train_vars(self):
        """
        must keep same order as grads calculated in fit_batch
        :return:
        """
        gen_vars = self.generator.trainable_variables
        dis_vars = self.discrimator.trainable_variables
        return gen_vars + dis_vars

    def fit_batch(self, inputs, targets=None):
        gen_vars = self.generator.trainable_variables
        dis_vars = self.discrimator.trainable_variables
        with tf.GradientTape() as gen_tape, tf.GradientTape() as dis_tape:
            outputs = self._model(inputs, training=True)
            losses = self.calc_loss(outputs, targets)
        gen_grads = self.calc_grads(gen_tape, gen_vars, losses['gen_loss'])
        dis_grads = self.calc_grads(dis_tape, dis_vars, losses['dis_loss'])
        grads = gen_grads + dis_grads
        self.apply_grads(gen_vars+dis_vars, grads)
        #opt.apply_gradients(zip(grads, gen_vars+dis_vars))
        return grads, losses


class ConstraintGANTrainer(object):
    def calc_loss(self, outputs, targets):
        losses = super(ConstraintGANTrainer, self).calc_loss(outputs, targets)
        if losses['fake_loss'] < self.cfg.min_fake_loss:
            losses['dis_loss'] = 0.0

        if losses['gen_loss'] < self.cfg.min_gen_loss:
            losses['gen_loss'] = 0.0
        return losses

if __name__ == "__main__":
    import logging
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s -   %(message)s', datefmt='%m/%d/%Y %H:%M:%S',
                        level=logging.INFO)
    import numpy as np
    def map_func(x):
        dis_real_targets = np.ones([x.numpy().shape[0]], np.float32)
        dis_fake_targets = 1-dis_real_targets
        gen_targets = dis_real_targets
        return x, (gen_targets, dis_real_targets, dis_fake_targets)

    ds1 = tf.data.Dataset.from_tensor_slices(np.random.rand(101, 128))
    ds2 = tf.data.Dataset.from_tensor_slices(np.random.rand(101, 28, 28, 1))
    ds = tf.data.Dataset.zip((ds1, ds2))
    ds = ds.map(lambda noise, img: {'img': img, 'noise':noise}).batch(5)

    cfg = GAN.cfg.copy()
    cfg.epochs = 3
    cfg.batch_size = 5
    cfg.accumulated_batch_size = 2
    cfg.run_eagerly = True
    model = GAN(cfg=cfg.__dict__)
    model.tffit(ds)







import numpy as np
import logging
import torch
from torch.optim.lr_scheduler import LambdaLR
import math
import mautil as mu

logger = logging.getLogger(__name__)

def tensor2device(tensor, device):
    if isinstance(tensor, list):
        return [tensor2device(t, device) for t in tensor]
    elif torch.is_tensor(tensor):
        return tensor.to(device)
    else:
        return tensor


def batch2device(batch, device):
    for k, v in batch.items():
        batch[k] = tensor2device(v, device)


def sequence_mask(lengths, maxlen=None, dtype=torch.bool):
    if maxlen is None:
        maxlen = lengths.max()
    row_vector = torch.arange(0, maxlen, 1, device=lengths.device)
    matrix = torch.unsqueeze(lengths, dim=-1)
    mask = row_vector < matrix
    mask.type(dtype)
    return mask


def get_linear_decay_schedule_with_warmup(optimizer, n_warmup_step, n_decay_step, decay_ratio=0.1, last_epoch=-1):
    """ Create a schedule with a learning rate that decreases following the
    values of the cosine function between 0 and `pi * cycles` after a warmup
    period during which it increases linearly between 0 and 1.
    """

    def lr_lambda(current_step):
        if current_step < n_warmup_step:
            ratio = float(current_step) / float(max(1, n_warmup_step))
        elif current_step<n_decay_step:
            progress = float(min(current_step, n_decay_step) - n_warmup_step) / (n_decay_step - n_warmup_step)

            ratio = 1-(1-decay_ratio)*progress
        else:
            ratio = decay_ratio
        return ratio

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def mesh_reduce_losses(losses):
    reduce_losses = {}
    for k in losses[0]:
        reduce_losses[k] = np.mean([loss[k] for loss in losses])
    return reduce_losses


def mesh_reduce_preds(preds):
    new_preds = {}
    for k in preds[0]:
        if isinstance(preds[0][k], list):
            new_preds[k] = [v for pred in preds for v in pred[k]]
        else:
            new_preds[k] = np.concat([pred[k] for pred in preds], 0)
    return new_preds


def mesh_reduce_outputs(outputs):
    return {k:torch.cat([output[k] for output in outputs], 0) for k in outputs[0]}



import numpy as np, time, logging, os
from queue import Queue
from threading import Thread
import multiprocessing as mp
from copy import deepcopy

from . import util

logger = logging.getLogger(__name__)

gl = globals()


class BasicBatchReader(object):
    def __init__(self, name, seed, data_dir, batch_process_func=None):
        self.name = name
        self._rs = np.random.RandomState(seed)
        self._batch_process_func = batch_process_func
        self._queue = None
        self.data_dir = data_dir
        self.state_fpath = os.path.join(data_dir, name + '.dump')
        self.state = {'rs': self._rs, 'batch_pos': 0}

    def set_rs(self, rs):
        self._rs = rs

    def save(self):
        state = deepcopy(self.state)
        if self._queue is not None:
            state['batch_pos'] = max(0, state['batch_pos'] - self._queue.qsize())

        util.dump(state, self.state_fpath)
        logger.info('batch reader %s batch position %s saved to file %s', self.name, state['batch_pos'],
                     self.state_fpath)

    def restore(self):
        if os.path.exists(self.state_fpath):
            self.state = util.load_dump(self.state_fpath)
        else:
            logger.warning('batch reader did not find the state file %s to restore, will do initial run',
                            self.state_fpath)
            self.state = {'rs': self._rs, 'batch_pos': 0}
        logger.info('batch reader %s current state is %s', self.name, self.state)


class BatchReader(BasicBatchReader):
    def __init__(self, name, seed, data_dir, model=None, batch_process_func=None, process_mode='T', sleep_seconds=1,
                 timeout=1, qsize=8, balanced_read=False, squeeze_seq=False, min_seq_len=-1, ordered=False, **kwargs):
        """

        :param name:
        :param seed:
        :param process_mode: T for thread, P for subprocess, N for neither
        :param sleep_seconds:
        :param timeout:
        :param qsize:
        :param squeeze_seq: seqs will be squeezed to max_seq_len of batch
        """
        super(BatchReader, self).__init__(name, seed, data_dir, batch_process_func)
        self._sleep_seconds = sleep_seconds
        self._process_mode = process_mode
        self._timeout = timeout
        self._qsize = qsize
        self._balanced_read = balanced_read
        self._do_squeeze_seq = squeeze_seq
        self._min_seq_len = min_seq_len
        self._ordered = ordered
        self._model = model
        self.kwargs = kwargs

        if self._process_mode == 'P':
            self._queue = mp.Queue(self._qsize)
        elif self._process_mode == 'T':
            self._queue = Queue(self._qsize)

    def _squeeze_seq(self, batch, seq_keys):
        for seq_key, seq_len_key in seq_keys:
            seq_key = '_'.join(seq_len_key.split('_')[:-1])
            max_len = max(self._min_seq_len, np.max(batch[seq_len_key]))
            batch[seq_key] = batch[seq_key][:, :max_len]
        return batch

    def _get_seqs_keys(self, keys):
        keys = set(keys)
        seqs_len_keys = []
        seqs_keys = []
        for key in keys:
            if key.endswith('seqs_len'):
                seqs_key = '_'.join(key.split('_')[:-1])
                if seqs_key in keys:
                    seqs_len_keys.append(key)
                    seqs_keys.append(seqs_key)
        return seqs_keys, seqs_len_keys

    def get_batch(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        self.state['rs'] = deepcopy(self._rs)
        batch_pos = self.state['batch_pos']
        if data_type != 'train':
            self.state['batch_pos'] = 0
        seqs_keys, seqs_len_keys = self._get_seqs_keys(x.keys())

        get_batch_func = self._get_batch
        if self._balanced_read and data_type == 'train':
            get_batch_func = self._get_batch_balanced
        for batch in get_batch_func(x, batch_size, y, shuffle, data_type):
            if self._do_squeeze_seq is True:
                batch = self._squeeze_seq(batch, zip(seqs_keys, seqs_len_keys))

            if self._batch_process_func is not None:
                batch = self._batch_process_func(batch, data_type=data_type)
            self.state['batch_pos'] = (self.state['batch_pos'] + 1) % self._num_batch
            yield batch
        if data_type == 'train':
            self.state['batch_pos'] = 0
        else:
            self.state['batch_pos'] = batch_pos
        self.state['rs'] = self._rs

    def _get_batch(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        for key in x:
            num = x[key].shape[0]
            break
        inds = np.arange(num)
        if shuffle and not self._ordered:
            self._rs.shuffle(inds)
        num_batch = (num + batch_size - 1) // batch_size
        self._num_batch = num_batch
        logger.info('batch reader %s: totally %s batchs', self.name, self._num_batch)
        batch_ids = np.arange(num_batch).astype(np.int32)
        if shuffle:
            self._rs.shuffle(batch_ids)

        for i in batch_ids[self.state['batch_pos']:]:
            batch = {}
            batch_inds = inds[batch_size * i:min(batch_size * (i + 1), num)]
            batch['batch_inds'] = batch_inds
            batch['batch_size'] = len(batch_inds)
            for key in x:
                batch[key] = x[key][batch_inds]

            if y is not None:
                for key in y:
                    batch[key] = y[key][batch_inds]
            yield batch

    def _gen_batch(self, x, batch_size, y, shuffle, data_type='train'):
        for i, batch in enumerate(self.get_batch(x, batch_size, y, shuffle=shuffle, data_type=data_type)):
            self._queue.put(batch, block=True)
        self._queue.put(None)

    def __call__(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        """

        :param x: dict of array. key name endswith _seqs, _seqs_len were preserved for seqs data
        :param batch_size:
        :param y:
        :param shuffle:
        :param data_type:
        :return:
        """

        if self._process_mode == 'P':
            thread = mp.Process(target=self._gen_batch, args=[x, batch_size, y, shuffle, data_type])
        elif self._process_mode == 'T':
            thread = Thread(target=self._gen_batch, args=[x, batch_size, y, shuffle, data_type])
        if self._process_mode in ['T', 'P']:
            thread.daemon = True
            thread.start()
            time.sleep(self._sleep_seconds)
            logger.info('name:%s, get batch sleep %s seconds', self.name, self._sleep_seconds)
            while True:
                batch = self._queue.get(True, self._timeout)
                if batch is None: break;
                yield batch
        elif self._process_mode == 'N':
            for batch in self.get_batch(x, batch_size, y, shuffle=shuffle, data_type=data_type):
                yield batch

    def _get_batch_balanced(self, x, batch_size, y=None, shuffle=True, data_type='train'):
        """
        In some case data is imbalance
        """

        label_inds = []
        labels = np.unique(x['labels'])
        min_num = np.inf
        for label in labels:
            label_ind = list(np.where(np.array(x['labels']) == label)[0])
            if shuffle and not self._ordered:
                self._rs.shuffle(label_ind)
            label_inds.append(label_ind)
            if len(label_ind) < min_num:
                min_num = len(label_ind)
        num_batch = (min_num + batch_size - 1) // batch_size
        self._num_batch = num_batch
        logger.info('batch reader %s: totally %s batchs', self.name, self._num_batch)
        batch_ids = np.arange(num_batch).astype(np.int32)
        if shuffle:
            self._rs.shuffle(batch_ids)

        for i in batch_ids[self.state['batch_pos']:]:
            batch = {}
            batch_inds = []
            for inds in label_inds:
                batch_inds += inds[batch_size * i:min(batch_size * (i + 1), min_num)]
            batch['batch_inds'] = batch_inds
            batch['batch_size'] = len(batch_inds)
            for key in x:
                batch[key] = x[key][batch_inds]

            if y is not None:
                for key in y:
                    batch[key] = y[key][batch_inds]

            yield batch